# Copyright (C) 2005, Timothy R. Morgan
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice and this permission
# notice appear in all copies, modified copies and in
# supporting documentation.

"React Model"

# This needs a bit of testing with different database types.
# Specifically, saving dates and retrieving auto increment ids is
# something that needs some attention. I don't have a lot of time
# to test nor a lot of knowledge about databases, so any help you
# can provide would be greatly appreciated.

__version__ = '0.2.1'

from Globals import InitializeClass, DTMLFile
from OFS.SimpleItem import SimpleItem
from AccessControl import ClassSecurityInfo
from OFS.PropertyManager import PropertyManager
from Shared.DC.ZRDB.Results import Results

class ReactModel(SimpleItem, PropertyManager):
  """
  A React Model handles the interaction between your application
  and its external SQL relational database tables.
  """

  meta_type = 'React Model'
  
  security = ClassSecurityInfo()
  
  manage_options = (
    PropertyManager.manage_options
    + SimpleItem.manage_options
  )

  _properties = (
    {'id': 'db_id', 'type': 'string', 'mode': 'w'},
    {'id': 'id_select_query', 'type': 'string', 'mode': 'w'},
    {'id': 'datetime_format', 'type': 'string', 'mode': 'w'},
  )
  
  def __init__(self, id, db_id, id_select_query, datetime_format):
    self.id = id
    self.db_id = db_id
    self.id_select_query = id_select_query
    self.datetime_format = datetime_format
  
  def get_db(self):
    """
    Returns the database connection object of the id
    specified by 'db_id'.
    """
    try:
      db = getattr(self.aq_parent, self.db_id)
    except:
      if self.db_id:
        raise Exception, 'Could not find database connection object "%s"' % self.db_id
      else:
        raise Exception, 'No database connection object specified.'
    return db._v_database_connection
  
  def new(self, **kwargs):
    """
    Creates a new React Model object. Properties on the object represent
    columns of the database record. Arguments passed to this method are set
    as propertis on the object.
    
    Examples::
    
      thing1 = models.things.new()
      thing1.name = "My First Thing"
      thing1.widgets = 24
      
      thing2 = models.things.new(name="My Second Thing", widgets=25)
    """
    return ReactModelRecord(self.get_db(), self.db_id, self.id_select_query, self.datetime_format, self.id, **kwargs)

  def find_all(self, conditions=None):
    """
    Finds and returns all records (React Model objects) matching 'conditions'
    (if specified)
    
    Arguments:
    
    conditions -- either a single string or a list of strings to match as
      the WHERE clause of the SQL query; if a list of strings is passed,
      they are joined together using "AND"
    """
    if conditions and type(conditions) == type([]):
      conditions = ' and '.join(conditions)
    if conditions:
      query = 'select id from %s where %s' % (self.id, conditions)
    else:
      query = 'select id from %s' % self.id
    db = self.get_db()
    # this is inefficient, but will work for now
    records = []
    for result in db.query(query)[1]:
      id = result[0]
      record = ReactModelRecord(self.get_db(), self.db_id, self.id_select_query, self.datetime_format, self.id)
      record.load_by_sql('id = %s' % id)
      records.append(record)
    return records
  
  def find(self, id):
    """
    Returns the React Model object with the id specified
    
    Arguments:
    
    id -- the integer id of the record
    """
    record = ReactModelRecord(self.get_db(), self.db_id, self.id_select_query, self.datetime_format, self.id)
    sql = 'id = %s' % id
    record.load_by_sql(sql)
    return record
  
  def find_by_sql(self, sql):
    """
    Using the SQL query string provided, query the database and return the
    first record (React Model object) found.
    
    Arguments:
    
    sql -- the SQL query string to use
    
    Example::
    
      thing = models.things.find_by_sql("name = 'My Thing'")
    """
    record = ReactModelRecord(self.get_db(), self.db_id, self.id_select_query, self.datetime_format, self.id)
    record.load_by_sql(sql)
    return record
  
  def find_by_value(self, column, value):
    """
    Find and return the record with 'column' matching 'value'.
    
    Arguments:
    
    column -- the column to lookup
      
    value -- the value to match
      
    Example::
    
      thing = models.things.find_by_value("name", "My Thing")
    """
    record = ReactModelRecord(self.get_db(), self.db_id, self.id_select_query, self.datetime_format, self.id)
    sql = "%s = %s" % (column, record._quote_value(value))
    record.load_by_sql(sql)
    return record


InitializeClass(ReactModel)

manage_addReactModelForm = DTMLFile('www/ReactModelAdd', globals())

def manage_addReactModel(self, id, db_id, id_select_query, datetime_format, REQUEST=None, submit=None):
  """Adds a new ReactModel object."""
  self._setObject(id, ReactModel(id, db_id, id_select_query, datetime_format))
  if REQUEST != None:
    REQUEST.RESPONSE.redirect(REQUEST['URL1']+'/manage_main')


class ReactModelRecord(object):
  """
  Represents a single database table record, i.e. React Model object.
  Properties on this object represent columns of the database table.
  Changing these properties do not affect the database record until
  the 'save()' method is called.
  """
  
  # If a Zope script/dtml/whatever object can get ahold of
  # one of these objects, then it should be able to do
  # whatever it wants, so give it everything.
  __allow_access_to_unprotected_subobjects__ = 1
  _guarded_writes = 1
  
  def __init__(self, db, db_id, id_select_query, datetime_format, table, **kwargs):
    self._db = db
    self._db_id = db_id
    self._id_select_query = id_select_query
    self._datetime_format = datetime_format
    self._table = table
    self.id = None
    # keyword arguments are treated as columns of the table
    # and added to the object
    for key in kwargs.keys():
      setattr(self, key, kwargs[key])
  
  def __guarded_setitem__(self, key, value):
    """
    allows setting of normal attributes (no underscores please)
    """
    if not key.startswith('_') and key != 'id':
      setattr(self, key, value)
  
  def load_by_sql(self, sql):
    """
    query with the sql provided, get the first record from the
    results, and append the columns as attributes to the object
    """
    try:
      results = self._db.query('select * from %s where %s' % (self._table, sql))
      record = Results(results).dictionaries()[0]
    except:
      raise Exception, 'Record could not be found.'
    else:
      for key in record.keys():
        setattr(self, key, record[key])
  
  def save(self):
    """
    Saves a React Model object (record) to the database table
    with which it is associated. If this object has an id,
    then do an update to an existing record; if not, insert
    a new record.
    """
    dict = self.__dict__.copy()
    for key in dict.keys():
      if key.startswith('_') or key == 'id':
        del(dict[key])
    columns = dict.keys()
    # UPDATE
    if self.id:
      sets = []
      for column in columns:
        value = self._quote_value(dict[column])
        sets.append('%s = %s' % (column, value))
      query = 'update %s set %s where id = %s' % (self._table, ', '.join(sets), self.id)
    # INSERT
    else:
      values = []
      for column in columns:
        value = self._quote_value(dict[column])
        values.append(value)
      query = 'insert into %s (%s) values (%s)' % (self._table, ', '.join(columns), ', '.join(values))
      if self._id_select_query:
        # Use the provided select query to get the last inserted id.
        query += '\0' + self._id_select_query
      else:
        # This is pretty ugly, but it should work in most situations.
        # This assumes that querying for the combination of all the values
        # (other than the id) and using the last record in the results
        # will do the trick.
        wheres = []
        for column in columns:
          value = self._quote_value(dict[column])
          wheres.append('%s = %s' % (column, value))
        query += "\0select id from %s where %s" % (self._table, ' and '.join(wheres))
      self.id = self._db.query(query)[1][-1][0]
  
  def destroy(self):
    """
    Deletes the record from the database. This operation is
    immediate and cannot be undone.
    """
    if self.id:
      query = 'delete from %s where id = %s' % (self._table, self.id)
      self._db.query(query)

  def _quote_value(self, value):
    """
    Return the properly stringified and quoted value based
    on the database type.
    """
    if type(value) == type(''):
      value = quote(value)
    elif hasattr(value, 'strftime'):
      value = quote(value.strftime(self._datetime_format))
    else:
      value = str(value)
    return value

def quote(value):
  """
  Escape quotes for a SQL query.
  """
  return "'" + value.replace("'", "''") + "'"
